@@ -1339,7 +1339,18 @@ class RegressionTest(RegressionMixin, jsonext.JSONSerializable):
         )
         exec_cmd = [self.job.launcher.run_command(self.job),
                     self.executable, *self.executable_opts]
-        commands = [*self.prerun_cmds, ' '.join(exec_cmd), *self.postrun_cmds]
+
+        if self.build_system:
+            prepare_cmds = self.build_system.prepare_cmds()
+        else:
+            prepare_cmds = []
+
+        commands = [
+            *prepare_cmds,
+            *self.prerun_cmds,
+            ' '.join(exec_cmd),
+            *self.postrun_cmds
+        ]
         user_environ = env.Environment(type(self).__name__,
                                        self.modules, self.variables.items())
         environs = [
